from __future__ import annotations

import base64
import importlib
import inspect
import json
import struct
from dataclasses import dataclass
from typing import BinaryIO, Dict, Iterable, TypedDict, Union

from pysll.decoders import decode

from .exceptions import VariableUnitValueUnparsableExpressionException
from .utils import constellation_date_from_string


@dataclass
class Object:
    id: str
    type: str = ""
    name: str | None = None
    local_path: str = ""

    def __str__(self):
        return self.sll_style_type()

    def __repr__(self):
        return self.__str__()

    def __eq__(self, o):
        if self.id != o.id:
            return False
        if self.type != "" and o.type != "":
            return self.type == o.type
        return True

    def sll_style_type(self):
        if self.type:
            type_string_without_object = self.type.replace("Object.", "")
            type_string = ", ".join(type_string_without_object.split("."))
            return f'Object[{type_string}, "{self.id}"]'
        return f'Object["{self.id}"]'


@dataclass
class Model:
    id: str
    type: str | None = None
    name: str | None = None

    def __eq__(self, other):
        return isinstance(other, Model) and (self.id, self.type or other.type) == (other.id, other.type or self.type)

    def sll_style_type(self) -> str:
        if not self.type:
            return f"Model[{self.id}]"

        parts = self.type.split(".")
        head, *types = parts
        if head != "Model":
            types = parts

        return f"Model[{', '.join(types + [self.id])}]"


Kind = Object | Model
ListableKind = Kind | Iterable[Kind]


@dataclass
class BlobRef:
    bucket: str
    path: str
    file_hash: str
    # The local path is set only on BlobRefs that have been downloaded locally
    local_path: str = ""

    def key(self):
        return self.path + self.file_hash

    def __str__(self):
        return "BlobRef[{}, {}, {}]".format(self.bucket, self.path, self.file_hash)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, o):
        return self.bucket == o.bucket and self.key() == o.key()

    @classmethod
    def from_dict(cls, data: dict):
        return BlobRef(data["bucket"], data["path"], data["hash"])


class ConstellationFieldDefinition:
    """A structure for holding field definitions from constellation field
    summaries."""

    def __init__(self, count, format, subfields):
        self.count = count
        self.format = format
        self.subfields = subfields

    @classmethod
    def from_dict(cls, data):
        """Create a ConstellationFieldDefinition from a dictionary."""
        return ConstellationFieldDefinition(
            data.get("Count"),
            data.get("Format"),
            [ConstellationSubFieldDefinition.from_dict(subfieldData) for subfieldData in data.get("SubFieldSummaries")],
        )

    def field_info(self, index=-1, name=""):
        """Return the info (e.g., type, units, etc.) for the supplied index and
        name."""
        for subfield in self.subfields:
            if index == -1 or subfield.index == index:
                if name == "" or subfield.name == name:
                    return subfield
        return None


class ConstellationSubFieldDefinition:
    """A structure for holding sub field definitions from constellation field
    summaries."""

    def __init__(self, ecl_type, unit="", name="", index=0):
        self.name = name
        self.index = index
        self.unit = unit
        self.ecl_type = ecl_type

    @classmethod
    def from_dict(cls, data):
        """Create a ConstellationSubFieldDefinition from a dictionary."""
        return ConstellationSubFieldDefinition(
            data.get("Type"),
            data.get("Unit"),
            data.get("Name"),
            data.get("Index"),
        )


class ConstellationFieldParser:
    """Responsible for parsing various constellation field types."""

    def __init__(self, blob_ref_download_api=None):
        self._field_summaries_by_type = {}
        self._cloud_files_encountered = []
        self._blob_refs_encountered = []
        self._blob_ref_download_api = blob_ref_download_api

    def add_summaries_for_type(self, summaries, ecl_type):
        """Add one or more summaries for fields of the supplied types.

        If there is an existing field summary, these will overwrite
        fields that match, but field summaries that are not passed here
        will be untouched in the existing summary store.
        """
        if ecl_type not in self._field_summaries_by_type:
            self._field_summaries_by_type[ecl_type] = {}
        for field_name, field_summary in summaries.items():
            self._field_summaries_by_type[ecl_type][field_name] = ConstellationFieldDefinition.from_dict(field_summary)

    def parse_field_value(self, ecl_type, field_name, field_value, traversal=None):
        """Parse the supplied field value into something that is python
        consumable."""

        # Start with trying to figure out the type from the summaries
        field_info = self._field_info_from_summary(ecl_type, field_name)
        if field_info:
            return self._parse_field_value(field_value, field_info, traversal=traversal)

        # If you couldn't read the summary, check the field value its self
        field_info = self._infer_field_type_from_value(field_value)
        if field_info:
            return self._parse_field_value(field_value, field_info, traversal=traversal)

        # If you couldn't infer it, then just return it unchanged
        return field_value

    def cloud_files_encountered(self):
        """Return a list of the cloud files that were encountered during the
        parsing."""
        return self._cloud_files_encountered

    def _field_info_from_summary(self, ecl_type, field_name):
        """Determine the field type (if possible) from the field summaries."""
        return self._field_summaries_by_type.get(ecl_type, {}).get(field_name)

    def _infer_field_type_from_value(self, field_value: Dict):
        """Try to infer the field type from the value."""
        subfield_values = []
        if isinstance(field_value, list):
            subfield_values = [subfield_value for subfield_value in field_value]
        else:
            subfield_values = [field_value]

        format = "Single"
        if len(subfield_values) > 1:
            # There's no way to infer the difference between a multiple field
            # and a named/indexed single field, so we assume its multiple
            format = "Multiple"

        subfield_definitions = []
        for subfield_value in subfield_values:
            if isinstance(subfield_value, dict):
                supplied_type = field_value.get("$Type")
                if supplied_type:
                    if supplied_type == "__JsonLink__":
                        subfield_definitions.append(ConstellationSubFieldDefinition("Link"))
                        continue
                    if supplied_type == "__JsonBlobReference__":
                        subfield_definitions.append(ConstellationSubFieldDefinition("BlobRef"))
                        continue
            subfield_definitions.append(ConstellationSubFieldDefinition("Unknown"))

        return ConstellationFieldDefinition(len(subfield_values), format, subfield_definitions)

    def _parse_field_value(self, field_value, field_info, map_children=True, traversal=None):
        """Parse the field value correctly based on the info."""
        # If we couldn't figure out the field info, then just return the value unchanged
        if not field_info or len(field_info.subfields) == 0:
            return field_value

        if field_value is None:
            return None

        if traversal and "columns" in traversal:
            if traversal["columns"]["start"] == traversal["columns"]["end"]:
                subfields = [field_info.subfields[traversal["columns"]["start"] - 1]]
            else:
                subfields = field_info.subfields[traversal["columns"]["start"] - 1 : traversal["columns"]["end"] - 1]
        else:
            subfields = field_info.subfields

        # If its a multiple field, then map over it (making sure to not infinitely recurse)
        if field_info.format == "Multiple" and map_children and isinstance(field_value, list):
            return [
                self._parse_field_value(subvalue, field_info, map_children=False, traversal=traversal)
                for subvalue in field_value
            ]

        if len(subfields) == 1:
            # If its a normal single field, then just return it parsed
            return self._parse_subfield(field_value, subfields[0])

        # If we're here, the subfield is named or indexed, so map the subfield values
        # to the subfield definitions
        if isinstance(field_value, list) and len(field_value) == len(subfields):
            results = []
            for subfield_value, subfield_info in zip(field_value, subfields):
                results.append(self._parse_subfield(subfield_value, subfield_info))
            return results

        # If we're here, then we weren't able to get the subfield definitions, so
        # return the value unchanged
        return field_value

    def _parse_subfield(self, field_value, field_info):
        """Parse a subfield value into something consumable by python."""

        # If field_value is None type, return without additional parsing
        if field_value is None:
            return field_value

        # Otherwise, see if we know how to process this type.
        if field_info.ecl_type == "Date":
            return constellation_date_from_string(field_value)
        if field_info.ecl_type == "Float":
            if field_info.unit:
                return VariableUnitValue(field_value, field_info.unit.replace('"', ""))
            return field_value
        if field_info.ecl_type == "String":
            return field_value
        if field_info.ecl_type == "Expression":
            return field_value
        if field_info.ecl_type == "Link":
            obj = self.to_link(field_value)
            if obj.type == "Object.EmeraldCloudFile":
                self._cloud_files_encountered.append(obj)
            return obj
        if field_info.ecl_type == "BlobRef":
            blob_ref = BlobRef.from_dict(field_value)
            self._blob_refs_encountered.append(blob_ref)
            if self._blob_ref_download_api:
                self._blob_ref_download_api(blob_ref)
                if blob_ref.local_path:
                    parsed_result = self.parse_local_file(blob_ref.local_path)
                    if parsed_result:
                        return parsed_result
            return blob_ref
        if field_info.ecl_type == "Byte":
            # If something is a byte field, its very likely to be a quantity array,
            # but not guaranteed. We'll need to investigate how to determine this 100%
            return self.parse_variable_unit_field(field_value)
        if field_info.ecl_type == "VariableUnit":
            return VariableUnitValue.parse_mathematica_expression(field_value.get("mathematica_value", ""))

        # If we don't, then just return the value unchanged
        return field_value

    def to_link(self, field_value):
        """Convert a dictionary field value to a link."""
        obj = field_value["object"]
        id, type = obj["id"], obj.get("type", "")
        klass = Model if type.startswith("Model") else Object
        return klass(id, type)

    def decompress_field(self, data):
        """Decompress the value of a compressed field."""

        return decode(data)

        # The data in a compressed field is first compressed with zlib, then base64 encoded,
        # and the proceeded with a 1. See:
        # https://mathematica.stackexchange.com/questions/104660/what-algorithm-do-the-compress-and-uncompress-functions-use for more info  # noqa: E501

    def parse_variable_unit_field(self, data):
        """Decompress and parse the value of a variable unit field."""
        return VariableUnitValue.parse_mathematica_expression(decode(data))

    def parse_local_file(self, local_path):
        """Attempt to parse a local file into python types.

        If this fails, it will return None
        """
        if not local_path:
            # If local file is not downloaded, it cannot be read
            return None

        # Blob refs are confusing because they are part ascii (specifically
        # a single line header) followed by binary data.  So we do something
        # weird where we read the file as binary and convert the first line
        # into ascii so we can read the header info
        with open(local_path, "rb") as file:
            # NOTE: this is relying on the convention that the header is always on a single line
            # e.g. ending with a newline character. If this ever changes in the blob ref encoding
            # scheme, then we have to edit the code below.
            header_byte_string = file.readline()

            header = header_byte_string.decode("utf-8")

            # convert the header into json
            header_data = json.loads(header)

            # Validate the header is something we can read
            if header_data.get("version") != 1 or header_data.get("encoder") != "binary":
                return None

            # Pull out the unit information, which is just base 64 encoded in the
            # header:
            units = base64.b64decode(header_data.get("mathematicaUnits"))
            parsed_units = VariableUnitValue.parse_mathematica_units_from_expression(units)

            # Parse the values in the file
            results = []
            for row_values in self._parse_binary_value(file, header_data.get("binaryRowFormat")):
                if len(row_values) == 0:
                    continue
                if len(row_values) != len(parsed_units or []):
                    return None
                row_results = []
                for value, unit in zip(row_values, parsed_units or []):
                    row_results.append(VariableUnitValue(value, unit))
                results.append(row_results)

        return results

    def _parse_binary_value(self, filestream, binary_row_format):
        """Parse binary values from a filestream according to the supplied
        binary row format."""

        # Figure out the data width and format based on the binary row format
        # The possible types are stored in the helpfile for ?BinaryRead in mathematica
        # under more options.  Basically, we need to define a map of how many bytes
        # each type is and what the corresponding python value is for each type
        data_width = 0  # size of the row in bytes
        data_format = "<"  # MM is always little endean
        for entry_format in binary_row_format:
            (
                entry_data_width,
                entry_format,
            ) = self._width_and_format_for_mathematica_binary_type(entry_format)
            if entry_format == "" or entry_data_width == 0:
                return None
            data_width += entry_data_width
            data_format += entry_format

        while True:
            data = filestream.read(data_width)
            if len(data) < data_width:
                # end of file
                return
            yield struct.unpack(data_format, data)

    def _width_and_format_for_mathematica_binary_type(self, entry_format):
        """Calculate the width (in bytes) and the correct python type for each
        mathematica type.

        See ?BinaryRead (under details and options) for more info on MM
        types. See
        https://docs.python.org/3/library/struct.html#format-characters
        for more details on python types
        """
        width_and_format_table = {
            "Byte": [1, "b"],
            "Character8": [1, "c"],
            "Character16": [2, ""],
            "Character32": [4, ""],
            "Complex64": [8, ""],
            "Complex128": [16, ""],
            "Complex256": [32, ""],
            "Integer8": [1, "b"],
            "Integer16": [2, "h"],
            "Integer24": [3, ""],
            "Integer32": [4, "i"],
            "Integer64": [8, "q"],
            "Integer128": [16, ""],
            "Real32": [4, "f"],
            "Real64": [8, "d"],
            "Real128": [16, ""],
            "TerminatedString": [0, ""],
            "UnsignedInteger8": [1, "B"],
            "UnsignedInteger16": [2, "H"],
            "UnsignedInteger24": [3, ""],
            "UnsignedInteger32": [4, "I"],
            "UnsignedInteger64": [8, "Q"],
            "UnsignedInteger128": [16, ""],
        }
        width, struct_format = width_and_format_table.get(entry_format, [0, ""])
        return width, struct_format


class VariableUnitValue:
    """Represents a value with a convertable unit (E.g., 1.2 mm)"""

    def __init__(self, value, unit):
        self.value = value
        self.unit = unit

    def __repr__(self):
        return str(self.value) + " " + self.unit

    def __eq__(self, o):
        return isinstance(o, VariableUnitValue) and self.value == o.value and self.unit == o.unit

    @classmethod
    def parse_mathematica_expression(cls, mm_expression):
        """Convert the provided MM expression into python structures."""
        return cls._recursively_parse_mm_expression(mm_expression)

    @classmethod
    def parse_mathematica_units_from_expression(cls, mm_expression):
        """Convert a mathematica expression into one or more units."""
        parsed_expression = cls._recursively_parse_mm_expression(mm_expression)
        return cls._parse_mathematica_units_from_parsed_expression(parsed_expression)

    @classmethod
    def _parse_mathematica_units_from_parsed_expression(cls, parsed_expression):
        """Recursively parse mathematica units from a parsed expression."""
        # Once you've hit a string, you have a real unit
        if isinstance(parsed_expression, str):
            return parsed_expression

        # Map over lists, being sure to accomadate "times" units
        if isinstance(parsed_expression, list):
            # Skip empty lists
            if len(parsed_expression) == 0:
                return None
            # Ignore the IndepdentUnit operation:
            if parsed_expression[0] == "IndependentUnit":
                return parsed_expression[1]

            # Perform the times operation
            if parsed_expression[0] == "Times":
                multiplied_unit = ""
                for subunit in parsed_expression[1:]:
                    parsed_subunit = cls._parse_mathematica_units_from_parsed_expression(subunit)
                    if parsed_subunit:
                        if multiplied_unit:
                            multiplied_unit += " "
                        multiplied_unit += parsed_subunit  # type:ignore
                return multiplied_unit

            # For normal lists, just map the result,
            # filtering None results and dropping the list keyword
            if parsed_expression[0] == "List":
                result = []
                for subunit in parsed_expression[1:]:
                    parsed_subunit = cls._parse_mathematica_units_from_parsed_expression(subunit)
                    if parsed_subunit:
                        result.append(parsed_subunit)
                return result

        return None

    @classmethod
    def _recursively_parse_mm_expression(cls, mm_expression):
        """Recursively parse a supplied mm expression into python
        structures."""
        if isinstance(mm_expression, bytes):
            mm_expression = mm_expression.decode("utf-8")

        mm_expression = mm_expression.strip()

        # Check if its the opening of a list
        if mm_expression.startswith("QuantityArray[") and mm_expression.endswith("]"):
            return cls._recursively_parse_mm_expression(mm_expression[14:-1])

        # Check if its the beginning of structured data
        if mm_expression.startswith("StructuredArray`StructuredData[") and mm_expression.endswith("]"):
            return cls._parse_structured_data_expression(mm_expression[31:-1])

        # Check if its the beginning of simple structured array
        if mm_expression.startswith("StructuredArray[") and mm_expression.endswith("]"):
            return cls._parse_structured_array_expression(mm_expression[16:-1])

        # Check if its a quanity
        if mm_expression.startswith("Quantity[") and mm_expression.endswith("]"):
            return cls._parse_quantity_expression(mm_expression[9:-1])

        # Check if its a rule list opening via List[Rule[ keyword
        if mm_expression.startswith("List[Rule[") and mm_expression.endswith("]"):
            return cls._parse_mm_rule_list_expression(mm_expression[5:-1])

        # Check if its a list opening via List keyword
        if mm_expression.startswith("List[") and mm_expression.endswith("]"):
            return cls._parse_mm_list_expression(mm_expression[5:-1])

        # Check if its a list opening via [ character
        if mm_expression.startswith("[") and mm_expression.endswith("]"):
            return cls._parse_mm_list_expression(mm_expression[1:-1])

        # Check if its a list opening via {}
        if mm_expression.startswith("{") and mm_expression.endswith("}"):
            return cls._parse_mm_list_expression(mm_expression[1:-1])

        # Check if its a Rule
        if mm_expression.startswith("Rule[") and mm_expression.endswith("]"):
            return cls._parse_mm_rule_expression(mm_expression[5:-1])

        # Check if its an object
        if mm_expression.startswith("Object[") and mm_expression.endswith("]"):
            return cls._parse_mm_object_expression(mm_expression[7:-1])

        # Check if its a MM null:
        if mm_expression == "Null":
            return None

        # Check if its an int
        try:
            return int(mm_expression.replace("`", ""))
        except ValueError:
            pass

        # Check if its a float
        try:
            return float(mm_expression.replace("`", ""))
        except ValueError:
            pass

        # At this point, its a symbol, so just return it as a string
        # we'll drop the enclosing quotes as they're not needed
        if mm_expression.startswith('"') and mm_expression.endswith('"'):
            mm_expression = mm_expression[1:-1]
        return mm_expression

    @classmethod
    def _parse_mm_list_expression(cls, mm_expression):
        """Parse a list mm expression into a python structure."""
        components = cls._split_by_unenclosed_substring(mm_expression, ",", {"{": "}", "[": "]"})
        return [cls._recursively_parse_mm_expression(component) for component in components]

    @classmethod
    def _parse_mm_rule_expression(cls, mm_expression):
        """Parse a rule mm expression into a python dict."""
        components = cls._split_by_unenclosed_substring(mm_expression, ",", {"{": "}", "[": "]"})
        if len(components) == 2:
            return {
                cls._recursively_parse_mm_expression(components[0]): cls._recursively_parse_mm_expression(  # type:ignore
                    components[1]
                )
            }
        return {}

    @classmethod
    def _parse_mm_rule_list_expression(cls, mm_expression):
        """Parse a rule list mm expression into a python dict."""
        components = cls._split_by_unenclosed_substring(mm_expression, ",", {"{": "}", "[": "]"})
        results = {}
        for component in components:
            result = cls._recursively_parse_mm_expression(component)
            if isinstance(result, dict):
                results.update(result)
        return results

    @classmethod
    def _parse_quantity_expression(cls, mm_expression):
        """Parse a quantity mm expression into a python structure."""
        components = cls._parse_mm_list_expression(mm_expression)
        if len(components) != 2:
            raise VariableUnitValueUnparsableExpressionException("Unexpected quantity components", components)

        return VariableUnitValue(components[0], components[1])

    @classmethod
    def _parse_mm_object_expression(cls, mm_expression):
        """Parse a list mm expression into a python structure."""
        components = cls._split_by_unenclosed_substring(mm_expression.replace(" ", ""), ",", {"{": "}", "[": "]"})
        return Object(id=components[-1], type=".".join(components[:-1]))

    @classmethod
    def _parse_structured_data_expression(cls, mm_expression):
        """Parse a structured data mm expression into a python structure."""
        # The structure data looks like:
        # {Length}, {{val1, ..., valn}, {unit1, ..., unitN}, {additional info}}
        # where each thing can also be a multiple if there's multiple values
        # we don't care about thet length or additional info,
        # so we're just going to pull the values and unit
        structured_data_components = cls._parse_mm_list_expression(mm_expression)

        # there's two forms (at least) that MM can put this in, so let's check each
        if len(structured_data_components) == 2:
            value_list = structured_data_components[1]
            if len(value_list) < 2:  # type:ignore
                raise VariableUnitValueUnparsableExpressionException("Unexpected value list", value_list)

            all_values = value_list[0]  # type:ignore
            all_units = value_list[1]  # type:ignore
        elif len(structured_data_components) > 2:
            all_values = structured_data_components[1]
            all_units = structured_data_components[2]
        else:
            raise VariableUnitValueUnparsableExpressionException(
                "Unexpected structured data components", structured_data_components
            )

        # Sometimes the values are at the top level, and sometimes they're one level
        # down - thanks MM!
        is_multiple = True
        if len(all_values) == 0 or not isinstance(all_values[0], list):  # type:ignore
            all_values = [all_values]
            is_multiple = False

        variable_unit_values = []
        for values in all_values:  # type:ignore
            # Convert the units and values into variable unit objects
            variable_unit_subvalues = []
            for value, unit in zip(values, all_units):  # type:ignore
                variable_unit_subvalues.append(VariableUnitValue(value, unit))
            if is_multiple:
                variable_unit_values.append(variable_unit_subvalues)
            else:
                variable_unit_values = variable_unit_subvalues
        return variable_unit_values

    @classmethod
    def _parse_structured_array_expression(cls, mm_expression):
        """Parse a structured array mm expression into a python structure."""
        components = cls._parse_mm_list_expression(mm_expression)
        if len(components) != 3:
            raise VariableUnitValueUnparsableExpressionException("Unexpected structured array components", components)

        return components[2]

    @classmethod
    def _find_unenclosed_substring(cls, string, substring, enclosing_chars):
        """Find, for example, a comma not contained in {}"""
        open_enclosing_chars = []
        for index in range(len(string)):
            if len(open_enclosing_chars) == 0 and string[index:].startswith(substring):
                return index
            for opening_char, closing_char in enclosing_chars.items():
                if string[index] == opening_char:
                    open_enclosing_chars.append(opening_char)
                if string[index] == closing_char:
                    open_enclosing_chars.remove(opening_char)
        return -1

    @classmethod
    def _split_by_unenclosed_substring(cls, string, substring, enclosing_chars):
        """Split, for example, a comma not contained in {}"""
        results = []
        index = cls._find_unenclosed_substring(string, substring, enclosing_chars)
        while index != -1:
            results.append(string[:index])
            string = string[index + 1 :]
            index = cls._find_unenclosed_substring(string, substring, enclosing_chars)
        results.append(string)
        return results


@dataclass
class CloudFileArgs:
    """Class for arguments that are needed to upload a cloud file to s3 and
    constellation."""

    file: BinaryIO
    name: str
    extension: str
    path: str
    size: int
    key: str
    content_md5: str
    content_disposition: str


class OptionItemDict(TypedDict):
    data_type: str
    name: str
    value: dict


@dataclass
class Option:
    name: str
    value: ListableType


@dataclass
class Quantity:
    value: float
    units: str


@dataclass
class Symbol:
    name: str


@dataclass
class Expression:
    string: str


class Function:
    def __init__(self, *args: ListableType, **kwargs: ListableType):
        self.inputs = list(args)
        self.options = self._build_options_from_kwargs(kwargs)

    def __repr__(self):
        # this isn't final, it just shows that the jupyter nb will render this nicely
        return type(self).__name__ + "[" + str(self.inputs)[:-1] + str(self.options) + "]"

    def _build_options_from_kwargs(self, kwargs: dict) -> list[Option]:
        return [Option(name, val) for name, val in kwargs.items()]

    @staticmethod
    def from_dict(d: dict) -> Function:
        """Method that converts a serialized dictionary into a Function."""
        try:
            assert d["data_type"] == "function"
        except (KeyError, AssertionError):
            raise TypeError("The input dictionary did not have the correct 'data_type' value.")

        try:
            members = {
                kname: klass
                for (kname, klass) in inspect.getmembers(importlib.import_module("pysll.functions"), inspect.isclass)
            }
            unit_op_members = {
                kname: klass
                for (kname, klass) in inspect.getmembers(
                    importlib.import_module("pysll.unit_operations"), inspect.isclass
                )
            }
            members.update(unit_op_members)
            klass = members[d["name"]]
            assert issubclass(klass, Function)
        except (KeyError, AssertionError):
            raise TypeError(
                "Name of the input function is not supported. Either the name has been mispelled or "
                "it is missing from this library."
            )

        return klass(
            *(deserialize_item(inp) for inp in d["inputs"]),
            **{o["name"]: deserialize_item(o["value"]) for o in d["options"]},
        )

    def payload(self) -> FunctionPayload:
        return FunctionPayload(
            data_type="function",
            name=type(self).__name__,
            inputs=serialize_item(self.inputs),
            options=serialize_item(self.options),
        )


@dataclass
class FunctionPayload:
    data_type: str
    name: str
    inputs: ListableDict
    options: ListableDict


@dataclass
class ResultPayload:
    result: ListableDict
    error: bool
    messages: list[str]


SLLItemType = Union[bool, int, float, str, Option, Quantity, Object, Function, Symbol, Expression]
ListableType = SLLItemType | list["ListableType"]
ListableDict = dict | list["ListableDict"]


# NOTE: had to add 'list[Option]' to the input to appease the python type checking gods
# it SHOULD work without it b/c ListableType contains list[Option] by construction
def serialize_item(input: ListableType | list[Option]) -> ListableDict:
    """Convenience function that works for atomic types like int, float, str,
    bool that just converts these to dicts before serializing to json's."""
    match input:
        case list():
            return [serialize_item(inp) for inp in input]
        case bool():
            return {"data_type": "bool", "value": input}
        case int():
            return {"data_type": "integer", "value": input}
        case float():
            return {"data_type": "float", "value": input}
        case str():
            return {"data_type": "string", "value": input}
        case Object(id=id, type=obj_type, name=name):
            return {"data_type": "object", "id": id, "type": obj_type, "name": name}
        # symbol is ultimately a convenience class that gets mapped to an expression on the backend
        case Symbol(name=name):
            return {"data_type": "expression", "string": name}
        case Expression(string=string):
            return {"data_type": "expression", "string": string}
        case Quantity(value=value, units=units):
            return {"data_type": "quantity", "value": value, "units": units}
        case Function(inputs=inputs, options=options):
            if not isinstance(inputs, list):
                raise ValueError("Function contains inputs that are not in a list format, which is prohibited")
            return {
                "data_type": "function",
                "name": type(input).__name__,
                "inputs": [serialize_item(inp) for inp in inputs],
                "options": [serialize_item(op) for op in options],
            }
        case Option(name=name, value=value):
            return {"data_type": "option", "name": name, "value": serialize_item(value)}
        case _:
            raise TypeError(f"The input with type, {type(input)}, cannot be serialized into a dict.")


def deserialize_item(input: ListableDict) -> ListableType:
    """Convenience function that converts a dictionary into a known object
    type."""
    if isinstance(input, list):
        return [deserialize_item(inp) for inp in input]
    data_type = input.get("data_type")
    match input["data_type"]:
        case "integer":
            return int(input["value"])
        case "float":
            return float(input["value"])
        case "string":
            return str(input["value"])
        case "bool":
            return bool(input["value"])
        case "symbol":
            return Symbol(input["name"])
        case "expression":
            return Expression(input["string"])
        case "object":
            return Object(id=input["id"], type=input["type"], name=input["name"])
        case "quantity":
            return Quantity(value=input["value"], units=input["units"])
        case "function":
            return Function.from_dict(input)
        case "option":
            return Option(name=input["name"], value=deserialize_item(input["value"]))
        case _:
            raise TypeError(f"Found a data type that cannot be deserialized: {data_type}")



import re
from typing import Any

from pysll.exceptions import (
    ConstellationObjectDoesNotExistException,
    TraversalTooDeepError,
)
from pysll.models import ConstellationFieldParser


def parse_results_from_traversals(objects_info, obj, traversals, field_parser, *, depth: int):
    """Parse out the response for the requested traversals."""

    if depth >= 100:
        raise TraversalTooDeepError

    object_info = objects_info.get(obj.id, {})
    results = []
    for traversal in traversals:
        if "name" not in traversal:
            continue
        field_name = traversal["name"]
        field_value = object_info.get(field_name)
        if "rows" in traversal and field_value is not None:
            if not isinstance(field_value, list):
                raise TypeError(f"Field {field_name} is not an array")

            start_idx = traversal["rows"]["start"] - 1
            end_idx = traversal["rows"]["end"] - 1

            try:
                if start_idx == end_idx:
                    field_value = field_value[start_idx]
                else:
                    field_value = field_value[start_idx : end_idx + 1]
            except IndexError:
                raise IndexError(f"Row index out of range for field {field_name}")

        if "columns" in traversal and field_value is not None:
            if not isinstance(field_value, list):
                raise TypeError(f"Field {field_name} is not an array")

            unwrap = False
            # Handle case where field_value is a single row after row indexing
            if "rows" in traversal and not isinstance(field_value[0], list):
                unwrap = True
                field_value = [field_value]

            start_idx = traversal["columns"]["start"] - 1
            end_idx = traversal["columns"]["end"] - 1

            try:
                if start_idx == end_idx:
                    field_value = [row[start_idx] for row in field_value]
                else:
                    field_value = [row[start_idx : end_idx + 1] for row in field_value]

                # Unwrap single row result
                if unwrap:
                    field_value = field_value[0]

            except IndexError:
                raise IndexError(f"Column index out of range for field {field_name}")

        if field_name is None or field_name == "All":
            all_dict = {}
            for field_name, field_value in object_info.items():
                all_dict[field_name] = field_parser.parse_field_value(
                    object_info.get("type"), field_name, field_value, traversal
                )
            return [all_dict]

        if "repeat" in traversal:
            if isinstance(field_value, list):
                subresults = []
                for field_obj in field_value:
                    subsubresults = []
                    # If we have more traversing to do to get the final field value,
                    # take the current level of recursion, find the final value, and append it
                    if "next" in traversal:
                        subtraversals = traversal["next"]
                        subsubresults.append(
                            parse_results_from_traversals(
                                objects_info,
                                field_parser.to_link(field_obj),
                                subtraversals,
                                field_parser,
                                depth=depth + 1,
                            )[0]
                        )
                    else:
                        subsubresults.append(
                            field_parser.parse_field_value(
                                object_info.get("type"),
                                field_name,
                                field_obj,
                                traversal,
                            )
                        )
                    subsubresults.append(
                        parse_results_from_traversals(
                            objects_info,
                            field_parser.to_link(field_obj),
                            traversals,
                            field_parser,
                            depth=depth + 1,
                        )[0]
                    )
                    subresults.append(subsubresults)
                results.append(subresults)
            elif field_value is not None:
                subresults = []
                if "next" in traversal:
                    subtraversals = traversal["next"]
                    subresults.append(
                        parse_results_from_traversals(
                            objects_info,
                            field_parser.to_link(field_value),
                            subtraversals,
                            field_parser,
                            depth=depth + 1,
                        )[0]
                    )
                else:
                    subresults.append(
                        field_parser.parse_field_value(
                            object_info.get("type"),
                            field_name,
                            field_value,
                            traversal,
                        )
                    )
                subsubresults = parse_results_from_traversals(
                    objects_info,
                    field_parser.to_link(field_value),
                    traversals,
                    field_parser,
                    depth=depth + 1,
                )[0]

                if subsubresults is not None:
                    subresults.extend(subsubresults)

                results.append(subresults)
            else:
                results.append(None)
        elif "next" in traversal:
            sub_traversals = traversal["next"]
            if isinstance(field_value, list):
                results.append(
                    [
                        parse_results_from_traversals(
                            objects_info,
                            field_parser.to_link(field_obj),
                            sub_traversals,
                            field_parser,
                            depth=depth + 1,
                        )[0]
                        for field_obj in field_value
                    ]
                )
            else:
                if field_value is None:
                    results.append(None)
                else:
                    results.append(
                        parse_results_from_traversals(
                            objects_info,
                            field_parser.to_link(field_value),
                            sub_traversals,
                            field_parser,
                            depth=depth + 1,
                        )[0]
                    )
        else:
            results.append(
                field_parser.parse_field_value(
                    object_info.get("type"),
                    field_name,
                    field_value,
                    traversal,
                )
            )

    return results


def parse_results_from_response(response_objects, objects, traversals, field_parser: ConstellationFieldParser):
    """Parse out the response for the requested fields and traversals."""
    results = []
    for obj in objects:
        obj_results = []
        for traversal in traversals:
            obj_results.append(
                parse_results_from_traversals(response_objects, obj, [traversal], field_parser, depth=0)[0]
            )
        results.append(obj_results)
    return results


def parse_field_values_from_response(response):
    """Parse out the field values contained in the response."""
    response_objects = {}
    field_summaries = {}
    response_fields = response.get("fields", {})
    response_field_summaries = response.get("summaries", {})
    ecl_type = response.get("resolved_object", {}).get("type")
    object_id = response.get("resolved_object", {}).get("id")

    if ecl_type is None or object_id is None:
        raise ConstellationObjectDoesNotExistException(response.get("object", {}).get("id"))

    response_objects[object_id] = {"type": ecl_type, "id": object_id}
    response_objects[object_id].update(response_fields)
    if ecl_type not in field_summaries:
        field_summaries[ecl_type] = {}
    field_summaries[ecl_type].update(response_field_summaries)
    for traversed_object in response.get("traversed_objects", []):
        (
            traversed_response_objects,
            traversed_field_summaries,
        ) = parse_field_values_from_response(traversed_object)
        response_objects.update(traversed_response_objects)

        # Merge field summaries instead of overwriting
        for type_key, summaries in traversed_field_summaries.items():
            if type_key not in field_summaries:
                field_summaries[type_key] = {}
            field_summaries[type_key].update(summaries)

    return response_objects, field_summaries


def build_traversal_from_tokens(tokens: list[str]):
    current_traversal = part_dict(tokens[0])

    if len(tokens) == 1:
        return current_traversal

    current_traversal.update({"next": [build_traversal_from_tokens(tokens[1:])]})
    return current_traversal


def build_traversal(field_string: str):
    tokens = split_on_brackets(field_string.replace(" ", ""))

    if len(tokens) == 0:
        return None

    return build_traversal_from_tokens(tokens)


def replace_repeated_pattern(input_string: str) -> str:
    pattern = r"Repeated\[(.*?)\]"
    return re.sub(pattern, r"{{\1}}", input_string)


def replace_array_pattern(input_string: str) -> str:
    # Match patterns like [[All,1]] or [[1]]
    pattern = r"\[\[(((All|\d+),)*\d+)\]\]"
    return re.sub(pattern, r"(\1)", input_string)


def split_on_brackets(input_string: str) -> list[str]:
    # First replace array access patterns
    transformed = replace_array_pattern(input_string)

    # Then replace Repeated[...] patterns
    transformed = replace_repeated_pattern(transformed)

    # Split on either '{{' or '['
    parts = [part for part in re.split(r"\[", transformed) if part]
    return [part.replace("]", "") for part in parts]


def part_dict(field_string: str) -> dict:
    single_index_pattern = r"(\w+)\((\d+)\)"
    multi_index_pattern = r"(\w+)\(All,\s*(\d+)\)"
    multi_specific_index_pattern = r"(\w+)\((\d+),\s*(\d+)\)"
    repeated_pattern = r"{{(.*)}}"

    repeated = False
    match = re.match(repeated_pattern, field_string)
    if match:
        repeated = True
        field_string = match.group(1)

    result: dict[str, Any] = {
        "name": field_string,
    }

    match = re.match(single_index_pattern, field_string)
    if match:
        result = {
            "name": match.group(1),
            "rows": {"start": int(match.group(2)), "end": int(match.group(2))},
        }

    match = re.match(multi_index_pattern, field_string)
    if match:
        result = {
            "name": match.group(1),
            "columns": {"start": int(match.group(2)), "end": int(match.group(2))},
        }

    match = re.match(multi_specific_index_pattern, field_string)
    if match:
        result = {
            "name": match.group(1),
            "rows": {"start": int(match.group(2)), "end": int(match.group(2))},
            "columns": {"start": int(match.group(3)), "end": int(match.group(3))},
        }

    if repeated:
        result["repeat"] = {"max": -1, "where": "", "inclusive_search": False}

    return result



from pysll.models import Function


# ------------------------------------ Manually defined operations -------------------------------------
class RoboticSamplePreparation(Function):
    """A wrapper that can be used to explicitly specify robotic execution of
    subsets of operations within a list of ExperimentSamplePreparation Unit
    Operations.

    The allowed input signatures of RoboticSamplePreparation is:

    ```RoboticSamplePreparation(UnitOperations)``` creates a `Protocol` object that will perform the robotic sample
    preparation specified in `UnitOperations`.

    :param `UnitOperations`: "The unit operations that specify how to perform the sample preparation.
    :returns: None - This is simply an inert head that is meant to separate manual and robotic preparation steps.

    All available unit operations can be found in the 'Experimental Principles' section of the site: https://www.emeraldcloudlab.com/helpfiles/experimentroboticsamplepreparation.
    """


class ManualSamplePreparation(Function):
    """A wrapper that can be used to explicitly specify manual execution of
    subsets of operations within a list of ExperimentSamplePreparation Unit
    Operations.

    The allowed input signatures of RoboticSamplePreparation is:

    ```ManualSamplePreparation(UnitOperations)``` creates a `Protocol` object that will perform the manual sample
    preparation specified in `UnitOperations`.

    :param `UnitOperations`: "The unit operations that specify how to perform the sample preparation.
    :returns: None - This is simply an inert head that is meant to separate manual and robotic preparation steps.

    All available unit operations can be found in the 'Experimental Principles' section of the site: https://www.emeraldcloudlab.com/helpfiles/experimentmanualsamplepreparation.
    """


# --------------------------- Automatically defined operations (paste below) ---------------------------
class AbsorbanceIntensity(Function):
    """The allowed input signatures of AbsorbanceIntensity without optional
    arguments (kwargs) are:

    ``AbsorbanceIntensity(Options, **kwargs)`` measures the absorbance intensity data of the input samples.

    :param `**kwargs`: optional arguments for AbsorbanceIntensity.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/absorbanceintensity.
    """


class AbsorbanceKinetics(Function):
    """The allowed input signatures of AbsorbanceKinetics without optional
    arguments (kwargs) are:

    ``AbsorbanceKinetics(Options, **kwargs)`` measures the absorbance kinetics data of the input samples.

    :param `**kwargs`: optional arguments for AbsorbanceKinetics.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/absorbancekinetics.
    """


class AbsorbanceSpectroscopy(Function):
    """The allowed input signatures of AbsorbanceSpectroscopy without optional
    arguments (kwargs) are:

    ``AbsorbanceSpectroscopy(Options, **kwargs)`` measures the absorbance spectroscopy data of the input samples.

    :param `**kwargs`: optional arguments for AbsorbanceSpectroscopy.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/absorbancespectroscopy.
    """


class AdjustpH(Function):
    """The allowed input signatures of AdjustpH without optional arguments
    (kwargs) are:

    ``AdjustpH(Options, **kwargs)`` adjusts the pHs of the given samples to the specified nominal pHs by adding acid and/or base.

    :param `**kwargs`: optional arguments for AdjustpH.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/adjustph.
    """


class Aliquot(Function):
    """The allowed input signatures of Aliquot without optional arguments
    (kwargs) are:

    ``Aliquot(aliquotRules, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes aliquoting of a single source to multiple destinations.

    :param `**kwargs`: optional arguments for Aliquot.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/aliquot.
    """


class AlphaScreen(Function):
    """The allowed input signatures of AlphaScreen without optional arguments
    (kwargs) are:

    ``AlphaScreen(Options, **kwargs)`` measures the alpha screen data of the input samples.

    :param `**kwargs`: optional arguments for AlphaScreen.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/alphascreen.
    """


class Centrifuge(Function):
    """The allowed input signatures of Centrifuge without optional arguments
    (kwargs) are:

    ``Centrifuge(centrifugeRules, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes centrifuging a sample at a specified intensity for a specified amount of time.

    :param `**kwargs`: optional arguments for Centrifuge.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/centrifuge.
    """


class ChangeMedia(Function):
    """The allowed input signatures of ChangeMedia without optional arguments
    (kwargs) are:

    ``ChangeMedia(changeMediaRules, **kwargs)`` generates an ExperimentRoboticCellPreparation-compatible 'primitive' that describes a change media process.

    :param `**kwargs`: optional arguments for ChangeMedia.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/changemedia.
    """


class CountLiquidParticles(Function):
    """The allowed input signatures of CountLiquidParticles without optional
    arguments (kwargs) are:

    ``CountLiquidParticles(Options, **kwargs)`` generates an ExperimentSamplePreparation-compatible 'UnitOperation' that run a high accuracy light obscuration (HIAC) experiment to count liquid particles of different sizes.

    :param `**kwargs`: optional arguments for CountLiquidParticles.
    :returns: UnitOperation - The unit operation that represents this measurement.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/countliquidparticles.
    """


class Cover(Function):
    """The allowed input signatures of Cover without optional arguments
    (kwargs) are:

    ``Cover(Sample, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes placing a lid onto a plate.

    :param `**kwargs`: optional arguments for Cover.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/cover.
    """


class CrossFlowFiltration(Function):
    """The allowed input signatures of CrossFlowFiltration without optional
    arguments (kwargs) are:

    ``CrossFlowFiltration(Options, **kwargs)`` generates an ExperimentSamplePreparation-compatible 'UnitOperation' that filter the provided sample by flowing it parallel-wise to the membrane surface.

    :param `**kwargs`: optional arguments for CrossFlowFiltration.
    :returns: UnitOperation - The unit operation that represents this measurement.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/crossflowfiltration.
    """


class Degas(Function):
    """The allowed input signatures of Degas without optional arguments
    (kwargs) are:

    ``Degas(Options, **kwargs)`` remove dissolved gases from the samples.

    :param `**kwargs`: optional arguments for Degas.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/degas.
    """


class Dilute(Function):
    """The allowed input signatures of Dilute without optional arguments
    (kwargs) are:

    ``Dilute(Options, **kwargs)`` transfers an amount of sample into a container and dilutes it with buffers as specified.

    :param `**kwargs`: optional arguments for Dilute.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/dilute.
    """


class DynamicLightScattering(Function):
    """The allowed input signatures of DynamicLightScattering without optional
    arguments (kwargs) are:

    ``DynamicLightScattering(Options, **kwargs)`` determine the hydrodynamic radius of an analyte via Dynamic Light Scattering.

    :param `**kwargs`: optional arguments for DynamicLightScattering.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/dynamiclightscattering.
    """


class ExtractPlasmidDNA(Function):
    """The allowed input signatures of ExtractPlasmidDNA without optional
    arguments (kwargs) are:

    ``ExtractPlasmidDNA(Options, **kwargs)`` isolates plasmid DNA from live cell or cell lysate through lysing (if dealing with cells, rather than lysate), then neutralizing the pH of the solution to keep plasmid DNA soluble (through renaturing) and pelleting out insoluble cell components, followed by one or more rounds of optional purification techniques including  precipitation (such as a cold ethanol or isopropanol wash), liquid-liquid extraction (such as phenol:chloroform extraction), solid phase extraction (such as spin columns), and magnetic bead separation (selectively binding plasmid DNA to magnetic beads while washing non-binding impurities from the mixture).

    :param `**kwargs`: optional arguments for ExtractPlasmidDNA.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/extractplasmiddna.
    """


class ExtractProtein(Function):
    """The allowed input signatures of ExtractProtein without optional
    arguments (kwargs) are:

    ``ExtractProtein(Options, **kwargs)`` isolates protein from live cell or cell lysate through lysing (if dealing with cells, rather than lysate), followed by one or more rounds of optional purification techniques including  precipitation (such as by adding ammonium sulfate, TCA (trichloroacetic acid), or acetone etc.), liquid-liquid extraction (e.g. adding C4 and C5 alcohols (butanol, pentanol) followed by ammonium sulfate into the protein-containing aqueous solution), solid phase extraction (such as spin columns), and magnetic bead separation (selectively binding proteins to magnetic beads while washing non-binding impurities from the mixture). Note that ExperimentExtractProtein is intended to extract specific or non-specific proteins from the whole cells or cell lysate.

    :param `**kwargs`: optional arguments for ExtractProtein.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/extractprotein.
    """


class ExtractRNA(Function):
    """The allowed input signatures of ExtractRNA without optional arguments
    (kwargs) are:

    ``ExtractRNA(Options, **kwargs)`` isolates RNA from live cell or cell lysate samples through lysing (if the input sample contains cells, rather than lysate), clearing the lysate of cellular debris by homogenization (optional), followed by one or more rounds of optional crude purification techniques including precipitation (such as a cold ethanol or isopropanol wash), liquid-liquid extraction (such as a phenol-chloroform extraction), solid phase extraction (such as a spin column), and magnetic bead separation (selectively binding RNA to magnetic beads while washing non-binding impurities from the mixture). Digestion enzymes can be added during any of these purification steps to degrade DNA in order to improve the purity of the extracted RNA. Extracted RNA can be further purified and analyzed with experiments including, but not limited to, ExperimentHPLC, ExperimentFPLC, and ExperimentPAGE (see experiment help files to learn more).

    :param `**kwargs`: optional arguments for ExtractRNA.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/extractrna.
    """


class FillToVolume(Function):
    """The allowed input signatures of FillToVolume without optional arguments
    (kwargs) are:

    ``FillToVolume(fillToVolumeRules, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes transfering a source into a destination until a desired volume is reached.

    :param `**kwargs`: optional arguments for FillToVolume.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/filltovolume.
    """


class Filter(Function):
    """The allowed input signatures of Filter without optional arguments
    (kwargs) are:

    ``Filter(Sample, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes filtering a sample through a specified filter by applying a specified pressure for a specified amount of time.

    :param `**kwargs`: optional arguments for Filter.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/filter.
    """


class FlashChromatography(Function):
    """The allowed input signatures of FlashChromatography without optional
    arguments (kwargs) are:

    ``FlashChromatography(Options, **kwargs)`` separate a sample via flash chromatography by flowing it through a column to which compounds in the sample will differentially adsorb.

    :param `**kwargs`: optional arguments for FlashChromatography.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/flashchromatography.
    """


class FlowCytometry(Function):
    """The allowed input signatures of FlowCytometry without optional arguments
    (kwargs) are:

    ``FlowCytometry(Options, **kwargs)`` flows a sample through a flow cytometer.

    :param `**kwargs`: optional arguments for FlowCytometry.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/flowcytometry.
    """


class FluorescenceIntensity(Function):
    """The allowed input signatures of FluorescenceIntensity without optional
    arguments (kwargs) are:

    ``FluorescenceIntensity(Options, **kwargs)`` measures the fluorescence intensity data of the input samples.

    :param `**kwargs`: optional arguments for FluorescenceIntensity.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/fluorescenceintensity.
    """


class FluorescenceKinetics(Function):
    """The allowed input signatures of FluorescenceKinetics without optional
    arguments (kwargs) are:

    ``FluorescenceKinetics(Options, **kwargs)`` measures the fluorescence kinetics data of the input samples.

    :param `**kwargs`: optional arguments for FluorescenceKinetics.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/fluorescencekinetics.
    """


class FluorescencePolarizationKinetics(Function):
    """The allowed input signatures of FluorescencePolarizationKinetics without
    optional arguments (kwargs) are:

    ``FluorescencePolarizationKinetics(Options, **kwargs)`` measures the fluorescence polarization kinetics data of the input samples.

    :param `**kwargs`: optional arguments for FluorescencePolarizationKinetics.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/fluorescencepolarizationkinetics.
    """


class FluorescencePolarization(Function):
    """The allowed input signatures of FluorescencePolarization without
    optional arguments (kwargs) are:

    ``FluorescencePolarization(Options, **kwargs)`` measures the fluorescence polarization data of the input samples.

    :param `**kwargs`: optional arguments for FluorescencePolarization.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/fluorescencepolarization.
    """


class FluorescenceSpectroscopy(Function):
    """The allowed input signatures of FluorescenceSpectroscopy without
    optional arguments (kwargs) are:

    ``FluorescenceSpectroscopy(Options, **kwargs)`` measures the fluorescence spectroscopy data of the input samples.

    :param `**kwargs`: optional arguments for FluorescenceSpectroscopy.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/fluorescencespectroscopy.
    """


class Grind(Function):
    """The allowed input signatures of Grind without optional arguments
    (kwargs) are:

    ``Grind(Options, **kwargs)`` reduces the size of powder particles by grinding solid substances into fine powders via a grinder (mill).

    :param `**kwargs`: optional arguments for Grind.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/grind.
    """


class GrowCrystal(Function):
    """The allowed input signatures of GrowCrystal without optional arguments
    (kwargs) are:

    ``GrowCrystal(Options, **kwargs)`` prepares crystallization plate designed to grow crystals, and incubate and image the prepared crystallization plate.

    :param `**kwargs`: optional arguments for GrowCrystal.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/growcrystal.
    """


class ICPMS(Function):
    """The allowed input signatures of ICPMS without optional arguments
    (kwargs) are:

    ``ICPMS(Options, **kwargs)`` atomize, ionize and analyze the elemental composition of the input analyte.

    :param `**kwargs`: optional arguments for ICPMS.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/icpms.
    """


class ImageCells(Function):
    """The allowed input signatures of ImageCells without optional arguments
    (kwargs) are:

    ``ImageCells(Options, **kwargs)`` acquire microscopic images from the samples.

    :param `**kwargs`: optional arguments for ImageCells.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/imagecells.
    """


class IncubateCells(Function):
    """The allowed input signatures of IncubateCells without optional arguments
    (kwargs) are:

    ``IncubateCells(Options, **kwargs)`` incubate cell samples for a specified period of time and at a specified temperature, humidity, and carbon dioxide percentage.

    :param `**kwargs`: optional arguments for IncubateCells.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/incubatecells.
    """


class Incubate(Function):
    """The allowed input signatures of Incubate without optional arguments
    (kwargs) are:

    ``Incubate(Sample, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes incubating and mixing a sample at a specified temperature and shaking rate for a specified amount of time.

    :param `**kwargs`: optional arguments for Incubate.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/incubate.
    """


class InoculateLiquidMedia(Function):
    """The allowed input signatures of InoculateLiquidMedia without optional
    arguments (kwargs) are:

    ``InoculateLiquidMedia(Options, **kwargs)`` moves bacterial colonies growing on solid media or in liquid media to fresh liquid media.

    :param `**kwargs`: optional arguments for InoculateLiquidMedia.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/inoculateliquidmedia.
    """


class LabelContainer(Function):
    """The allowed input signatures of LabelContainer without optional
    arguments (kwargs) are:

    ``LabelContainer(Options, **kwargs)`` generates an ExperimentSamplePreparation/ExperimentCellPreparation-compatible 'UnitOperation' that labels a container for use in other primitives.

    :param `**kwargs`: optional arguments for LabelContainer.
    :returns: UnitOperation - The primitive that represents this labeled container.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/labelcontainer.
    """


class LabelSample(Function):
    """The allowed input signatures of LabelSample without optional arguments
    (kwargs) are:

    ``LabelSample(Options, **kwargs)`` generates an ExperimentSamplePreparation/ExperimentCellPreparation-compatible 'UnitOperation' that labels a sample in a container for use in other primitives.

    :param `**kwargs`: optional arguments for LabelSample.
    :returns: UnitOperation - The primitive that represents this labeled sample.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/labelsample.
    """


class LiquidLiquidExtraction(Function):
    """The allowed input signatures of LiquidLiquidExtraction without optional
    arguments (kwargs) are:

    ``LiquidLiquidExtraction(Options, **kwargs)`` separates the aqueous and organic phases of given samples via pipette or phase separator, in order to isolate a target analyte that is more concentrated in either the aqueous or organic phase.

    :param `**kwargs`: optional arguments for LiquidLiquidExtraction.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/liquidliquidextraction.
    """


class LuminescenceIntensity(Function):
    """The allowed input signatures of LuminescenceIntensity without optional
    arguments (kwargs) are:

    ``LuminescenceIntensity(Options, **kwargs)`` measures the luminescence intensity data of the input samples.

    :param `**kwargs`: optional arguments for LuminescenceIntensity.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/luminescenceintensity.
    """


class LuminescenceKinetics(Function):
    """The allowed input signatures of LuminescenceKinetics without optional
    arguments (kwargs) are:

    ``LuminescenceKinetics(Options, **kwargs)`` measures the luminescence kinetics data of the input samples.

    :param `**kwargs`: optional arguments for LuminescenceKinetics.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/luminescencekinetics.
    """


class LuminescenceSpectroscopy(Function):
    """The allowed input signatures of LuminescenceSpectroscopy without
    optional arguments (kwargs) are:

    ``LuminescenceSpectroscopy(Options, **kwargs)`` measures the luminescence spectroscopy data of the input samples.

    :param `**kwargs`: optional arguments for LuminescenceSpectroscopy.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/luminescencespectroscopy.
    """


class LyseCells(Function):
    """The allowed input signatures of LyseCells without optional arguments
    (kwargs) are:

    ``LyseCells(Options, **kwargs)`` ruptures the cell membranes of a cell containing sample to enable extraction of cellular components.

    :param `**kwargs`: optional arguments for LyseCells.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/lysecells.
    """


class MagneticBeadSeparation(Function):
    """The allowed input signatures of MagneticBeadSeparation without optional
    arguments (kwargs) are:

    ``MagneticBeadSeparation(Options, **kwargs)`` isolates targets from samples by using a magnetic field to separate superparamagnetic particles from suspensions.

    :param `**kwargs`: optional arguments for MagneticBeadSeparation.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/magneticbeadseparation.
    """


class MeasureContactAngle(Function):
    """The allowed input signatures of MeasureContactAngle without optional
    arguments (kwargs) are:

    ``MeasureContactAngle(Options, **kwargs)`` measure the contact angle between fiber and wetting liquid.

    :param `**kwargs`: optional arguments for MeasureContactAngle.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/measurecontactangle.
    """


class MeasureRefractiveIndex(Function):
    """The allowed input signatures of MeasureRefractiveIndex without optional
    arguments (kwargs) are:

    ``MeasureRefractiveIndex(Options, **kwargs)`` measure the refractive index of sample.

    :param `**kwargs`: optional arguments for MeasureRefractiveIndex.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/measurerefractiveindex.
    """


class MicrowaveDigestion(Function):
    """The allowed input signatures of MicrowaveDigestion without optional
    arguments (kwargs) are:

    ``MicrowaveDigestion(Options, **kwargs)`` digest sample with microwave to ensure full solubility for subsequent analysis, especially for ICP-MS.

    :param `**kwargs`: optional arguments for MicrowaveDigestion.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/microwavedigestion.
    """


class Mix(Function):
    """The allowed input signatures of Mix without optional arguments (kwargs)
    are:

    ``Mix(Sample, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes mixing a sample using bench-top instrumentation or by pipetting on a micro liquid handling robot.

    :param `**kwargs`: optional arguments for Mix.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/mix.
    """


class NephelometryKinetics(Function):
    """The allowed input signatures of NephelometryKinetics without optional
    arguments (kwargs) are:

    ``NephelometryKinetics(Options, **kwargs)`` measures the nephelometry kinetics data of the input samples.

    :param `**kwargs`: optional arguments for NephelometryKinetics.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/nephelometrykinetics.
    """


class Nephelometry(Function):
    """The allowed input signatures of Nephelometry without optional arguments
    (kwargs) are:

    ``Nephelometry(Options, **kwargs)`` measures the nephelometry data of the input samples.

    :param `**kwargs`: optional arguments for Nephelometry.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/nephelometry.
    """


class PCR(Function):
    """The allowed input signatures of PCR without optional arguments (kwargs)
    are:

    ``PCR(Options, **kwargs)`` amplifies target sequences from nucleic acid samples.

    :param `**kwargs`: optional arguments for PCR.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/pcr.
    """


class Pellet(Function):
    """The allowed input signatures of Pellet without optional arguments
    (kwargs) are:

    ``Pellet(pelletRules, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes pelleting a sample at a specified intensity for a specified amount of time.

    :param `**kwargs`: optional arguments for Pellet.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/pellet.
    """


class PickColonies(Function):
    """The allowed input signatures of PickColonies without optional arguments
    (kwargs) are:

    ``PickColonies(Options, **kwargs)`` moves bacterial colonies growing on solid media to a new liquid or solid media.

    :param `**kwargs`: optional arguments for PickColonies.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/pickcolonies.
    """


class Precipitate(Function):
    """The allowed input signatures of Precipitate without optional arguments
    (kwargs) are:

    ``Precipitate(Options, **kwargs)`` combines precipitating reagent with sample and separates the resulting precipitate and liquid phase.

    :param `**kwargs`: optional arguments for Precipitate.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/precipitate.
    """


class Resuspend(Function):
    """The allowed input signatures of Resuspend without optional arguments
    (kwargs) are:

    ``Resuspend(resuspendRules, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes the resuspension of a sample in solvent.

    :param `**kwargs`: optional arguments for Resuspend.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/resuspend.
    """


class SerialDilute(Function):
    """The allowed input signatures of SerialDilute without optional arguments
    (kwargs) are:

    ``SerialDilute(Options, **kwargs)`` transfers an amount of sample into a container and dilutes it with buffers as specified.

    :param `**kwargs`: optional arguments for SerialDilute.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/serialdilute.
    """


class SolidPhaseExtraction(Function):
    """The allowed input signatures of SolidPhaseExtraction without optional
    arguments (kwargs) are:

    ``SolidPhaseExtraction(Sample, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes solid phase extraction a sample through a specified extraction cartridge by applying a specified pressure for a specified amount of time.

    :param `**kwargs`: optional arguments for SolidPhaseExtraction.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/solidphaseextraction.
    """


class SpreadCells(Function):
    """The allowed input signatures of SpreadCells without optional arguments
    (kwargs) are:

    ``SpreadCells(Options, **kwargs)`` moves suspended colonies growing in liquid media to solid media and moves them across the surface of the media in a pattern to promote growth of the colonies.

    :param `**kwargs`: optional arguments for SpreadCells.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/spreadcells.
    """


class StreakCells(Function):
    """The allowed input signatures of StreakCells without optional arguments
    (kwargs) are:

    ``StreakCells(Options, **kwargs)`` moves suspended colonies growing in liquid media to solid media and moves them across the surface of the media in a pattern to try and isolate individual colonies.

    :param `**kwargs`: optional arguments for StreakCells.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/streakcells.
    """


class ThawCells(Function):
    """The allowed input signatures of ThawCells without optional arguments
    (kwargs) are:

    ``ThawCells(Options, **kwargs)`` thaw a frozen vial of cells for use in downstream cell culturing.

    :param `**kwargs`: optional arguments for ThawCells.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/thawcells.
    """


class Transfer(Function):
    """The allowed input signatures of Transfer without optional arguments
    (kwargs) are:

    ``Transfer(transferRules, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes a transfer of a source to a destination.

    :param `**kwargs`: optional arguments for Transfer.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/transfer.
    """


class Uncover(Function):
    """The allowed input signatures of Uncover without optional arguments
    (kwargs) are:

    ``Uncover(Sample, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes removing a lid from a plate.

    :param `**kwargs`: optional arguments for Uncover.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/uncover.
    """


class VisualInspection(Function):
    """The allowed input signatures of VisualInspection without optional
    arguments (kwargs) are:

    ``VisualInspection(Options, **kwargs)`` records a video of a sample container as it is agitated on a shaker/vortex.

    :param `**kwargs`: optional arguments for VisualInspection.
    :returns: UnitOperation - The unit operation that is to be used in sample/cell preparation.

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/visualinspection.
    """


class Wait(Function):
    """The allowed input signatures of Wait without optional arguments (kwargs)
    are:

    ``Wait(duration, **kwargs)`` generates an ExperimentSampleManipulation-compatible 'primitive' that describes the pausing of a specified duration.

    :param `**kwargs`: optional arguments for Wait.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/wait.
    """


class WashCells(Function):
    """The allowed input signatures of WashCells without optional arguments
    (kwargs) are:

    ``WashCells(washCellsRules, **kwargs)`` generates an ExperimentRoboticCellPreparation-compatible 'primitive' that describes a wash cells process.

    :param `**kwargs`: optional arguments for WashCells.
    :returns: None

    All optional parameters (kwargs) can be found in the 'Options' tab of the site: https://www.emeraldcloudlab.com/helpfiles/washcells.
    """
